@@ -4,7 +4,6 @@
 # author: Thomas Moreau and Olivier Grisel
 #
 # adapted from concurrent/futures/process_pool_executor.py (17/02/2017)
-#  * Backport for python2.7/3.3,
 #  * Add an extra management thread to detect executor_manager_thread failures,
 #  * Improve the shutdown process to avoid deadlocks,
 #  * Add timeout for workers,
@@ -62,36 +61,29 @@ __author__ = 'Thomas Moreau (thomas.moreau.2010@gmail.com)'
 import os
 import gc
 import sys
+import queue
 import struct
 import weakref
 import warnings
 import itertools
 import traceback
 import threading
-from time import time
+from time import time, sleep
 import multiprocessing as mp
 from functools import partial
 from pickle import PicklingError
+from concurrent.futures import Executor
+from concurrent.futures._base import LOGGER
+from concurrent.futures.process import BrokenProcessPool as _BPPException
+from multiprocessing.connection import wait
 
-from . import _base
+from ._base import Future
 from .backend import get_context
-from .backend.compat import queue
-from .backend.compat import wait
-from .backend.compat import set_cause
 from .backend.context import cpu_count
 from .backend.queues import Queue, SimpleQueue
 from .backend.reduction import set_loky_pickler, get_loky_pickler_name
-from .backend.utils import recursive_terminate, get_exitcodes_terminated_worker
-
-try:
-    from concurrent.futures.process import BrokenProcessPool as _BPPException
-except ImportError:
-    _BPPException = RuntimeError
-
-
-# Compatibility for python2.7
-if sys.version_info[0] == 2:
-    ProcessLookupError = OSError
+from .backend.utils import kill_process_tree, get_exitcodes_terminated_worker
+from .initializers import _prepare_initializer
 
 
 # Mechanism to prevent infinite process spawning. When a worker of a
@@ -115,7 +107,9 @@ try:
         if force_gc:
             gc.collect()
 
-        return Process(pid).memory_info().rss
+        mem_size = Process(pid).memory_info().rss
+        mp.util.debug(f'psutil return memory size: {mem_size}')
+        return mem_size
 
 except ImportError:
     _USE_PSUTIL = False
@@ -134,12 +128,7 @@ class _ThreadWakeup:
 
     def wakeup(self):
         if not self._closed:
-            if sys.platform == "win32" and sys.version_info[:2] < (3, 4):
-                # Compat for python2.7 on windows, where poll return false for
-                # b"" messages. Use the slightly larger message b"0".
-                self._writer.send_bytes(b"0")
-            else:
-                self._writer.send_bytes(b"")
+            self._writer.send_bytes(b"")
 
     def clear(self):
         if not self._closed:
@@ -147,7 +136,7 @@ class _ThreadWakeup:
                 self._reader.recv_bytes()
 
 
-class _ExecutorFlags(object):
+class _ExecutorFlags:
     """necessary references to maintain executor states without preventing gc
 
     It permits to keep the information needed by executor_manager_thread
@@ -195,8 +184,9 @@ def _python_exit():
     global _global_shutdown
     _global_shutdown = True
     items = list(_threads_wakeups.items())
-    mp.util.debug("Interpreter shutting down. Waking up "
-                  "executor_manager_thread {}".format(items))
+    if len(items) > 0:
+        mp.util.debug("Interpreter shutting down. Waking up "
+                      f"executor_manager_thread {items}")
     for _, (shutdown_lock, thread_wakeup) in items:
         with shutdown_lock:
             thread_wakeup.wakeup()
@@ -224,7 +214,7 @@ class _RemoteTraceback(Exception):
     """Embed stringification of remote traceback in local traceback
     """
     def __init__(self, tb=None):
-        self.tb = '\n"""\n{}"""'.format(tb)
+        self.tb = f'\n"""\n{tb}"""'
 
     def __str__(self):
         return self.tb
@@ -246,11 +236,11 @@ class _ExceptionWithTraceback(BaseException):
 
 
 def _rebuild_exc(exc, tb):
-    exc = set_cause(exc, _RemoteTraceback(tb))
+    exc.__cause__ = _RemoteTraceback(tb)
     return exc
 
 
-class _WorkItem(object):
+class _WorkItem:
 
     __slots__ = ["future", "fn", "args", "kwargs"]
 
@@ -261,7 +251,7 @@ class _WorkItem(object):
         self.kwargs = kwargs
 
 
-class _ResultItem(object):
+class _ResultItem:
 
     def __init__(self, work_id, exception=None, result=None):
         self.work_id = work_id
@@ -269,7 +259,7 @@ class _ResultItem(object):
         self.result = result
 
 
-class _CallItem(object):
+class _CallItem:
 
     def __init__(self, work_id, fn, args, kwargs):
         self.work_id = work_id
@@ -285,8 +275,9 @@ class _CallItem(object):
         return self.fn(*self.args, **self.kwargs)
 
     def __repr__(self):
-        return "CallItem({}, {}, {}, {})".format(
-            self.work_id, self.fn, self.args, self.kwargs)
+        return (
+            f"CallItem({self.work_id}, {self.fn}, {self.args}, {self.kwargs})"
+        )
 
 
 class _SafeQueue(Queue):
@@ -296,7 +287,7 @@ class _SafeQueue(Queue):
         self.thread_wakeup = thread_wakeup
         self.pending_work_items = pending_work_items
         self.running_work_items = running_work_items
-        super(_SafeQueue, self).__init__(max_size, reducers=reducers, ctx=ctx)
+        super().__init__(max_size, reducers=reducers, ctx=ctx)
 
     def _on_queue_feeder_error(self, e, obj):
         if isinstance(obj, _CallItem):
@@ -310,8 +301,7 @@ class _SafeQueue(Queue):
                     "Could not pickle the task to send it to the workers.")
             tb = traceback.format_exception(
                 type(e), e, getattr(e, "__traceback__", None))
-            raised_error = set_cause(raised_error,
-                                     _RemoteTraceback(''.join(tb)))
+            raised_error.__cause__ = _RemoteTraceback(''.join(tb))
             work_item = self.pending_work_items.pop(obj.work_id, None)
             self.running_work_items.remove(obj.work_id)
             # work_item can be None if another process terminated. In this
@@ -322,15 +312,12 @@ class _SafeQueue(Queue):
                 del work_item
             self.thread_wakeup.wakeup()
         else:
-            super(_SafeQueue, self)._on_queue_feeder_error(e, obj)
+            super()._on_queue_feeder_error(e, obj)
 
 
 def _get_chunks(chunksize, *iterables):
     """Iterates over zip()ed iterables in chunks. """
-    if sys.version_info < (3, 3):
-        it = itertools.izip(*iterables)
-    else:
-        it = zip(*iterables)
+    it = zip(*iterables)
     while True:
         chunk = tuple(itertools.islice(it, chunksize))
         if not chunk:
@@ -374,8 +361,8 @@ def _process_worker(call_queue, result_queue, initializer, initargs,
             to by the worker.
         initializer: A callable initializer, or None
         initargs: A tuple of args for the initializer
-        process_management_lock: A ctx.Lock avoiding worker timeout while some
-            workers are being spawned.
+        processes_management_lock: A ctx.Lock avoiding worker timeout while
+            some workers are being spawned.
         timeout: maximum time to wait for a new item in the call_queue. If that
             time is expired, the worker will shutdown.
         worker_exit_lock: Lock to avoid flagging the executor as broken on
@@ -386,7 +373,7 @@ def _process_worker(call_queue, result_queue, initializer, initargs,
         try:
             initializer(*initargs)
         except BaseException:
-            _base.LOGGER.critical('Exception in initializer:', exc_info=True)
+            LOGGER.critical('Exception in initializer:', exc_info=True)
             # The parent will notice that the process stopped and
             # mark the pool broken
             return
@@ -398,15 +385,14 @@ def _process_worker(call_queue, result_queue, initializer, initargs,
     _last_memory_leak_check = None
     pid = os.getpid()
 
-    mp.util.debug('Worker started with timeout=%s' % timeout)
+    mp.util.debug(f'Worker started with timeout={timeout}')
     while True:
         try:
             call_item = call_queue.get(block=True, timeout=timeout)
             if call_item is None:
                 mp.util.info("Shutting down worker on sentinel")
         except queue.Empty:
-            mp.util.info("Shutting down worker after timeout %0.3fs"
-                         % timeout)
+            mp.util.info(f"Shutting down worker after timeout {timeout:0.3f}s")
             if processes_management_lock.acquire(block=False):
                 processes_management_lock.release()
                 call_item = None
@@ -421,12 +407,23 @@ def _process_worker(call_queue, result_queue, initializer, initargs,
                 # If we cannot format correctly the exception, at least print
                 # the traceback.
                 print(previous_tb)
+            mp.util.debug('Exiting with code 1')
             sys.exit(1)
         if call_item is None:
-            # Notify queue management thread about clean worker shutdown
+            # Notify queue management thread about worker shutdown
             result_queue.put(pid)
-            with worker_exit_lock:
-                return
+            is_clean = worker_exit_lock.acquire(True, timeout=30)
+
+            # Early notify any loky executor running in this worker process
+            # (nested parallelism) that this process is about to shutdown to
+            # avoid a deadlock waiting undifinitely for the worker to finish.
+            _python_exit()
+
+            if is_clean:
+                mp.util.debug('Exited cleanly')
+            else:
+                mp.util.info('Main process did not release worker_exit')
+            return
         try:
             r = call_item()
         except BaseException as e:
@@ -467,11 +464,12 @@ def _process_worker(call_queue, result_queue, initializer, initargs,
                 mp.util.info("Memory leak detected: shutting down worker")
                 result_queue.put(pid)
                 with worker_exit_lock:
+                    mp.util.debug('Exit due to memory leak')
                     return
         else:
             # if psutil is not installed, trigger gc.collect events
             # regularly to limit potential memory leaks due to reference cycles
-            if ((_last_memory_leak_check is None) or
+            if (_last_memory_leak_check is None or
                     (time() - _last_memory_leak_check >
                      _MEMORY_LEAK_CHECK_DELAY)):
                 gc.collect()
@@ -508,8 +506,12 @@ class _ExecutorManagerThread(threading.Thread):
         def weakref_cb(_,
                        thread_wakeup=self.thread_wakeup,
                        shutdown_lock=self.shutdown_lock):
-            mp.util.debug('Executor collected: triggering callback for'
-                          ' QueueManager wakeup')
+            if mp is not None:
+                # At this point, the multiprocessing module can already be
+                # garbage collected. We only log debug info when still
+                # possible.
+                mp.util.debug('Executor collected: triggering callback for'
+                              ' QueueManager wakeup')
             with shutdown_lock:
                 thread_wakeup.wakeup()
 
@@ -542,7 +544,7 @@ class _ExecutorManagerThread(threading.Thread):
         # of new processes or shut down
         self.processes_management_lock = executor._processes_management_lock
 
-        super(_ExecutorManagerThread, self).__init__()
+        super().__init__(name="ExecutorManagerThread")
         if sys.version_info < (3, 9):
             self.daemon = True
 
@@ -619,7 +621,7 @@ class _ExecutorManagerThread(threading.Thread):
                         "A task has failed to un-serialize. Please ensure that"
                         " the arguments of the function are all picklable."
                     )
-                    set_cause(bpe, result_item)
+                    bpe.__cause__ = result_item
                 else:
                     is_broken = False
             except BaseException as e:
@@ -630,7 +632,7 @@ class _ExecutorManagerThread(threading.Thread):
                 )
                 tb = traceback.format_exception(
                     type(e), e, getattr(e, "__traceback__", None))
-                set_cause(bpe,  _RemoteTraceback(''.join(tb)))
+                bpe.__cause__ = _RemoteTraceback(''.join(tb))
 
         elif wakeup_reader in ready:
             # This is simply a wake-up event that might either trigger putting
@@ -644,14 +646,21 @@ class _ExecutorManagerThread(threading.Thread):
                 # In Windows, introspecting terminated workers exitcodes seems
                 # unstable, therefore they are not appended in the exception
                 # message.
-                exit_codes = "\nThe exit codes of the workers are {}".format(
-                    get_exitcodes_terminated_worker(self.processes))
+                exit_codes = (
+                    "\nThe exit codes of the workers are "
+                    f"{get_exitcodes_terminated_worker(self.processes)}"
+                )
+            mp.util.debug('A worker unexpectedly terminated. Workers that '
+                          'might have caused the breakage: '
+                          + str({p.name: p.exitcode
+                                 for p in list(self.processes.values())
+                                 if p is not None and p.sentinel in ready}))
             bpe = TerminatedWorkerError(
                 "A worker process managed by the executor was unexpectedly "
                 "terminated. This could be caused by a segmentation fault "
                 "while calling the function or by an excessive memory usage "
                 "causing the Operating System to kill the worker.\n"
-                "{}".format(exit_codes)
+                f"{exit_codes}"
             )
 
         self.thread_wakeup.clear()
@@ -669,9 +678,12 @@ class _ExecutorManagerThread(threading.Thread):
             with self.processes_management_lock:
                 p = self.processes.pop(result_item, None)
 
-            # p can be None is the executor is concurrently shutting down.
+            # p can be None if the executor is concurrently shutting down.
             if p is not None:
                 p._worker_exit_lock.release()
+                mp.util.debug(
+                    f"joining {p.name} when processing {p.pid} as result_item"
+                )
                 p.join()
                 del p
 
@@ -690,7 +702,8 @@ class _ExecutorManagerThread(threading.Thread):
                         "executor. This can be caused by a too short worker "
                         "timeout or by a memory leak.", UserWarning
                     )
-                    executor._adjust_process_count()
+                    with executor._processes_management_lock:
+                        executor._adjust_process_count()
                     executor = None
         else:
             # Received a _ResultItem so mark the future as completed.
@@ -725,7 +738,7 @@ class _ExecutorManagerThread(threading.Thread):
         self.executor_flags.flag_as_broken(bpe)
 
         # Mark pending tasks as failed.
-        for work_id, work_item in self.pending_work_items.items():
+        for work_item in self.pending_work_items.values():
             work_item.future.set_exception(bpe)
             # Delete references to object. See issue16284
             del work_item
@@ -733,7 +746,7 @@ class _ExecutorManagerThread(threading.Thread):
 
         # Terminate remaining workers forcibly: the queues or their
         # locks may be in a dirty state and block forever.
-        self.kill_workers()
+        self.kill_workers(reason="broken executor")
 
         # clean up resources
         self.join_executor_internals()
@@ -753,17 +766,17 @@ class _ExecutorManagerThread(threading.Thread):
                 del work_item
 
             # Kill the remaining worker forcibly to no waste time joining them
-            self.kill_workers()
+            self.kill_workers(reason="executor shutting down")
 
-    def kill_workers(self):
+    def kill_workers(self, reason=''):
         # Terminate the remaining workers using SIGKILL. This function also
         # terminates descendant workers of the children in case there is some
         # nested parallelism.
         while self.processes:
             _, p = self.processes.popitem()
-            mp.util.debug('terminate process {}'.format(p.name))
+            mp.util.debug(f"terminate process {p.name}, reason: {reason}")
             try:
-                recursive_terminate(p)
+                kill_process_tree(p)
             except ProcessLookupError:  # pragma: no cover
                 pass
 
@@ -777,22 +790,36 @@ class _ExecutorManagerThread(threading.Thread):
         with self.processes_management_lock:
             n_children_to_stop = 0
             for p in list(self.processes.values()):
+                mp.util.debug(f"releasing worker exit lock on {p.name}")
                 p._worker_exit_lock.release()
                 n_children_to_stop += 1
 
+        mp.util.debug(f"found {n_children_to_stop} processes to stop")
+
         # Send the right number of sentinels, to make sure all children are
         # properly terminated. Do it with a mechanism that avoid hanging on
         # Full queue when all workers have already been shutdown.
         n_sentinels_sent = 0
+        cooldown_time = 0.001
         while (n_sentinels_sent < n_children_to_stop
                 and self.get_n_children_alive() > 0):
-            for i in range(n_children_to_stop - n_sentinels_sent):
+            for _ in range(n_children_to_stop - n_sentinels_sent):
                 try:
                     self.call_queue.put_nowait(None)
                     n_sentinels_sent += 1
-                except queue.Full:
+                except queue.Full as e:
+                    if cooldown_time > 10.0:
+                        raise e
+                    mp.util.info(
+                        "full call_queue prevented to send all sentinels at "
+                        "once, waiting..."
+                    )
+                    sleep(cooldown_time)
+                    cooldown_time *= 2
                     break
 
+        mp.util.debug(f"sent {n_sentinels_sent} sentinels to the call queue")
+
     def join_executor_internals(self):
         self.shutdown_workers()
 
@@ -814,13 +841,23 @@ class _ExecutorManagerThread(threading.Thread):
             self.thread_wakeup.close()
 
         # If .join() is not called on the created processes then
-        # some ctx.Queue methods may deadlock on Mac OS X.
-        mp.util.debug("joining processes")
-        for p in list(self.processes.values()):
-            p.join()
+        # some ctx.Queue methods may deadlock on macOS.
+        with self.processes_management_lock:
+            mp.util.debug(f"joining {len(self.processes)} processes")
+            n_joined_processes = 0
+            while True:
+                try:
+                    pid, p = self.processes.popitem()
+                    mp.util.debug(f"joining process {p.name} with pid {pid}")
+                    p.join()
+                    n_joined_processes += 1
+                except KeyError:
+                    break
 
-        mp.util.debug("executor management thread clean shutdown of worker "
-                      "processes: {}".format(list(self.processes)))
+            mp.util.debug(
+                "executor management thread clean shutdown of "
+                f"{n_joined_processes} workers"
+            )
 
     def get_n_children_alive(self):
         # This is an upper bound on the number of children alive.
@@ -834,9 +871,8 @@ _system_limited = None
 
 def _check_system_limits():
     global _system_limits_checked, _system_limited
-    if _system_limits_checked:
-        if _system_limited:
-            raise NotImplementedError(_system_limited)
+    if _system_limits_checked and _system_limited:
+        raise NotImplementedError(_system_limited)
     _system_limits_checked = True
     try:
         nsems_max = os.sysconf("SC_SEM_NSEMS_MAX")
@@ -851,8 +887,10 @@ def _check_system_limits():
         # minimum number of semaphores available
         # according to POSIX
         return
-    _system_limited = ("system provides too few semaphores (%d available, "
-                       "256 necessary)" % nsems_max)
+    _system_limited = (
+        f"system provides too few semaphores ({nsems_max} available, "
+        "256 necessary)"
+    )
     raise NotImplementedError(_system_limited)
 
 
@@ -880,8 +918,8 @@ def _check_max_depth(context):
     if 0 < MAX_DEPTH and _CURRENT_DEPTH + 1 > MAX_DEPTH:
         raise LokyRecursionError(
             "Could not spawn extra nested processes at depth superior to "
-            "MAX_DEPTH={}. If this is intendend, you can change this limit "
-            "with the LOKY_MAX_DEPTH environment variable.".format(MAX_DEPTH))
+            f"MAX_DEPTH={MAX_DEPTH}. If this is intendend, you can change "
+            "this limit with the LOKY_MAX_DEPTH environment variable.")
 
 
 class LokyRecursionError(RuntimeError):
@@ -918,7 +956,7 @@ class ShutdownExecutorError(RuntimeError):
     """
 
 
-class ProcessPoolExecutor(_base.Executor):
+class ProcessPoolExecutor(Executor):
 
     _at_exit = None
 
@@ -947,8 +985,7 @@ class ProcessPoolExecutor(_base.Executor):
             initargs: A tuple of arguments to pass to the initializer.
             env: A dict of environment variable to overwrite in the child
                 process. The environment variables are set before any module is
-                loaded. Note that this only works with the loky context and it
-                is unreliable under windows with Python < 3.6.
+                loaded. Note that this only works with the loky context.
         """
         _check_system_limits()
 
@@ -964,11 +1001,9 @@ class ProcessPoolExecutor(_base.Executor):
         self._context = context
         self._env = env
 
-        if initializer is not None and not callable(initializer):
-            raise TypeError("initializer must be a callable")
-        self._initializer = initializer
-        self._initargs = initargs
-
+        self._initializer, self._initargs = _prepare_initializer(
+            initializer, initargs
+        )
         _check_max_depth(self._context)
 
         if result_reducers is None:
@@ -1035,17 +1070,6 @@ class ProcessPoolExecutor(_base.Executor):
         if self._executor_manager_thread is None:
             mp.util.debug('_start_executor_manager_thread called')
 
-            # When the executor gets garbarge collected, the weakref callback
-            # will wake up the queue management thread so that it can terminate
-            # if there is no pending work item.
-            def weakref_cb(
-                    _, thread_wakeup=self._executor_manager_thread_wakeup,
-                    shutdown_lock=self._shutdown_lock):
-                mp.util.debug('Executor collected: triggering callback for'
-                              ' QueueManager wakeup')
-                with self._shutdown_lock:
-                    thread_wakeup.wakeup()
-
             # Start the processes so that their sentinels are known.
             self._executor_manager_thread = _ExecutorManagerThread(self)
             self._executor_manager_thread.start()
@@ -1070,7 +1094,7 @@ class ProcessPoolExecutor(_base.Executor):
                         _python_exit)
 
     def _adjust_process_count(self):
-        for _ in range(len(self._processes), self._max_workers):
+        while len(self._processes) < self._max_workers:
             worker_exit_lock = self._context.BoundedSemaphore(1)
             args = (self._call_queue, self._result_queue, self._initializer,
                     self._initargs, self._processes_management_lock,
@@ -1086,7 +1110,10 @@ class ProcessPoolExecutor(_base.Executor):
             p._worker_exit_lock = worker_exit_lock
             p.start()
             self._processes[p.pid] = p
-        mp.util.debug('Adjust process count : {}'.format(self._processes))
+        mp.util.debug(
+            f"Adjusted process count to {self._max_workers}: "
+            f"{[(p.name, pid) for pid, p in self._processes.items()]}"
+        )
 
     def _ensure_executor_running(self):
         """ensures all workers and management thread are running
@@ -1110,7 +1137,7 @@ class ProcessPoolExecutor(_base.Executor):
                 raise RuntimeError('cannot schedule new futures after '
                                    'interpreter shutdown')
 
-            f = _base.Future()
+            f = Future()
             w = _WorkItem(f, fn, args, kwargs)
 
             self._pending_work_items[self._queue_count] = w
@@ -1121,7 +1148,7 @@ class ProcessPoolExecutor(_base.Executor):
 
             self._ensure_executor_running()
             return f
-    submit.__doc__ = _base.Executor.submit.__doc__
+    submit.__doc__ = Executor.submit.__doc__
 
     def map(self, fn, *iterables, **kwargs):
         """Returns an iterator equivalent to map(fn, iter).
@@ -1150,13 +1177,14 @@ class ProcessPoolExecutor(_base.Executor):
         if chunksize < 1:
             raise ValueError("chunksize must be >= 1.")
 
-        results = super(ProcessPoolExecutor, self).map(
+        results = super().map(
             partial(_process_chunk, fn), _get_chunks(chunksize, *iterables),
-            timeout=timeout)
+            timeout=timeout
+        )
         return _chain_from_iterable_of_lists(results)
 
     def shutdown(self, wait=True, kill_workers=False):
-        mp.util.debug('shutting down executor %s' % self)
+        mp.util.debug(f'shutting down executor {self}')
 
         self._flags.flag_as_shutting_down(kill_workers)
         executor_manager_thread = self._executor_manager_thread
@@ -1178,4 +1206,4 @@ class ProcessPoolExecutor(_base.Executor):
         self._result_queue = None
         self._processes_management_lock = None
 
-    shutdown.__doc__ = _base.Executor.shutdown.__doc__
+    shutdown.__doc__ = Executor.shutdown.__doc__
